/**************************************************************************************************
* NOTICE OF COPYRIGHT AND OWNERSHIP OF SOFTWARE:
*
* Copyright 2010 Flexera Software, Inc. and/or InstallShield Co. Inc.  All Rights Reserved.  
*
* This computer program is the property of Flexera Software, Inc. and InstallShield Co. Inc. of 
* Schaumburg, Illinois, U.S.A.  Any use, copy, publication, distribution, display, modification, 
* or transmission of this computer program in whole or in part in any form or by any means without
* the prior express written permission of Flexera Software, Inc. or InstallShield Co. Inc. is 
* strictly prohibited.
* 
* Except when expressly provided by Flexera Software, Inc. or InstallShield Co. Inc. in writing,
* possession of this computer program shall not be construed to confer any license or rights under 
* any of Flexera Software, Inc.s or InstallShield Co. Inc.s intellectual property rights, whether
* by estoppel, implication, or otherwise.  
* 
* ALL COPIES OF THIS PROGRAM MUST DISPLAY THIS NOTICE OF COPYRIGHT AND OWNERSHIP IN FULL.
**************************************************************************************************/
/*
	This source file is part of a sample utility that demonstrates Composite Transactions.

	It defines the functions that relate to the FlxCtuContext (context) object. 
*/
#include "compTranUtilContext.h"

static FlxActBool sFindCommand(const FlxCtuContext*	pContext, 
							   const char*			pszCommandName, 
							   size_t*				pCommandIndex);

static FlxActBool sGetFirstRequestAction(const FlxCtuContext*	pContext, 
										 const FlxCtuCommand**	ppCommand);

/*******************************************************************************
*	flxCtuContextCreate
*
* Create a new command context and return a pointer to it.
* The FlxCtuCommandDef list defines the commands that are allowed. 
*******************************************************************************/
void flxCtuContextCreate(const FlxCtuCommandDef*	pCommandDefs, 
						 size_t						commandDefCount,
						 FlxCtuContext**			ppContext)
{
	*ppContext = (FlxCtuContext*)flxCtuMallocAndZeroise(sizeof(FlxCtuContext));
	(*ppContext)->exitCode			= flxCtuExitUnknown;
	(*ppContext)->pCommandDefs		= pCommandDefs;
	(*ppContext)->commandDefCount	= commandDefCount;
}

void flxCtuContextDestroy(FlxCtuContext* pContext)
{
	if ((pContext != (FlxCtuContext*)0))
	{
		/* Destroy the commands.	*/
		size_t commandIndex;
		for (commandIndex = 0; commandIndex < pContext->commandCount; commandIndex++)
			flxCtuCommandDestroy(pContext->ppCommands[commandIndex]);
		
		/* Free the list. */
		flxCtuFree(pContext->ppCommands);
		
		/* Free the strings. */
		flxCtuFree(pContext->pszProgramName);
		
		/* Destroy the command definitions. */
		flxCtuCommandDefsDestroy((FlxCtuCommandDef*)pContext->pCommandDefs, 
								 pContext->commandDefCount);

		/* Finally free the context structure itself. */
		flxCtuFree(pContext);
	}
}

const FlxCtuCommand* flxCtuContextGetCommand(const FlxCtuContext* pContext, size_t commandIndex)
{
	FLX_CTU_ASSERT( commandIndex < pContext->commandCount );
	return pContext->ppCommands[commandIndex];
}

/***************************************************************************************************
*		flxCtuContextGetCommandByName
*
*	Get a pointer to the first FlxCtuCommand object with the specified name. 
*	Return true if pCommandName is present in the command list.
*	Specify the command name in lower case.
*	Set the command index in the context so that the next command can be obtained by calling
*	flxCtuCommandGetNextCommand
***************************************************************************************************/
FlxActBool flxCtuContextGetCommandByName(const FlxCtuContext*	pContext, 
										 const char*			pszCommandName, 
										 const FlxCtuCommand**	ppCommand)
{
	size_t commandIndex;

	if ( sFindCommand(pContext, pszCommandName, &commandIndex) )
	{
		*ppCommand = pContext->ppCommands[commandIndex];
		return FLX_ACT_TRUE;
	}
	return FLX_ACT_FALSE;
}

size_t flxCtuContextGetCommandCount(const FlxCtuContext* pContext)
{
	return pContext->commandCount;
}

/***************************************************************************************************
*		flxCtuContextFileDump
*
*	Create/replace a file from a buffer.
***************************************************************************************************/
FlxActBool flxCtuContextFileDump(FlxCtuContext*	pContext, 
								 const char*	pszFilename, 
								 const char*	pFileContent,
								 size_t			contentSize)
{
	FILE*	hFile;

	hFile = fopen(pszFilename, "w");
	if (hFile == NULL)
	{
		flxCtuPrintError("Can't create file: %s\n", pszFilename);
		pContext->exitCode = flxCtuExitInputError;	/* Assume incorrect path given. */
		return FLX_ACT_FALSE;
	}
	if (1 != fwrite(pFileContent, contentSize, 1, hFile))
	{
		fclose(hFile);
		flxCtuPrintError("Can't write file: %s\n", pszFilename);
		pContext->exitCode = flxCtuExitFileError;
		return FLX_ACT_FALSE;
	}
	fclose(hFile);
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*		flxCtuContextFileLoad
*
*	Load a file into memory (free with flxCtuFree).
***************************************************************************************************/
FlxActBool flxCtuContextFileLoad(FlxCtuContext*	pContext, 
								 const char*	pszFilename, 
								 char**			ppFileContent,
								 size_t*		pContentSize)
{
	FILE*				hFile;
	FlxActBool			res;

	/* Open the file.	*/
	hFile = fopen(pszFilename, "r");
	if (hFile  == NULL)
	{
		flxCtuPrintError("can't open file %s\n", pszFilename);
		pContext->exitCode = flxCtuExitInputError;	/* Assume incorrect filename given. */
		return FLX_ACT_FALSE;
	}

	res = flxCtuContextStreamLoad(pContext, hFile, pszFilename, ppFileContent, pContentSize);

	fclose(hFile);

	return res;
}

/***************************************************************************************************
*		flxCtuContextStreamLoad
*
*	Load the contents of a stream into memory (free with flxCtuFree).
***************************************************************************************************/
FlxActBool flxCtuContextStreamLoad(FlxCtuContext*	pContext, 
								FILE*          	pFile,
							 	const char*		pszFilename, 
								char**			ppFileContent,
								size_t*			pContentSize)
{
	size_t	fileSize=0;
	char*	contents=NULL;

	/* Check the file stream.	*/
	if (pFile == NULL)
	{
		flxCtuPrintError("can't access file %s\n", pszFilename);
		pContext->exitCode = flxCtuExitInputError;	/* Assume incorrect filename given. */
		return FLX_ACT_FALSE;
	}

	/* Read from the stream.  */
	while( !ferror(pFile) && !feof(pFile) )
	{
		char	buffer[256];			
		size_t	actual;
	
		actual=fread(buffer, 1, sizeof(buffer), pFile);

		if( actual > 0 )
		{
			size_t newSize = fileSize + actual;
			contents = flxCtuRealloc(contents, newSize); 
			memcpy( &contents[fileSize], buffer, actual );
			fileSize = newSize;
		}
	}

	/* Check reading terminated successfully.  */
	if( !feof(pFile) )
	{
		flxCtuPrintError("Can't read file %s\n", pszFilename);
		pContext->exitCode = flxCtuExitFileError;
		flxCtuFree(contents);
		return FLX_ACT_FALSE;
	}
	
	/* Check for empty file. */
	if (fileSize == 0)
	{
		flxCtuPrintError("File %s is empty.\n", pszFilename);
		pContext->exitCode = flxCtuExitFileError;
		flxCtuFree(contents);
		return FLX_ACT_FALSE;
	}

	/* Allocate extra byte of memory for NUL terminator. */
	contents = flxCtuRealloc(contents, fileSize+1); 

	/* Append a NUL as a convenience for the caller. */
	contents[fileSize] = '\0';

	/* Return file stream contents plus the size if required. */
	*ppFileContent = contents;	
	if (pContentSize != NULL)
		*pContentSize = fileSize;

	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*		flxCtuContextFindCommandDef
*
***************************************************************************************************/
FlxActBool flxCtuContextFindCommandDef(const FlxCtuContext*		pContext, 
									   const char*				pszComamndOrAlias,
									   const FlxCtuCommandDef**	ppCommandDef)
{
	const char*	pszComamndOrAliasLwr = flxCtuStrDupLwr(pszComamndOrAlias);
	size_t index;
	for (index = 0; index < pContext->commandDefCount; index++)
	{
		if (	(strcmp(pszComamndOrAliasLwr, pContext->pCommandDefs[index].pszName) == 0)
			|| 	(strcmp(pszComamndOrAliasLwr, pContext->pCommandDefs[index].pszAlias) == 0) )
		{
			*ppCommandDef = &pContext->pCommandDefs[index];
			break;
		}
	}
	flxCtuFree(pszComamndOrAliasLwr);

	return (index < pContext->commandDefCount)? FLX_ACT_TRUE : FLX_ACT_FALSE;
}

FlxActBool flxCtuCommandHasTrait(const FlxCtuCommand*	pCommand,
								 FlxCtuCommandTrait		trait)
{
	return (pCommand->pDef->traits & trait) != 0;
}

/***************************************************************************************************
*		flxCtuContextHasCommand
*
*	Return true if pCommandName is present in the command list.
*	Specify the command name in lower case.
***************************************************************************************************/
FlxActBool flxCtuContextHasCommand(const FlxCtuContext* pContext, const char* pszCommandName)
{
	return sFindCommand(pContext, pszCommandName, NULL);
}

/***************************************************************************************************
*		flxCtuContextPrintApiFlexError
*
* These are errors from non- flxActTranXXX functions that report errors using a FlxActError object.
***************************************************************************************************/
void flxCtuContextPrintApiFlexError(FlxCtuContext* pContext, FlxActError* pFlxError)
{
	flxCtuPrintError("API function failed (%u,%u,%u)\n",
					 pFlxError->majorErrorNo,
					 pFlxError->minorErrorNo,
					 pFlxError->sysErrorNo);
}

/***************************************************************************************************
 *		flxCtuContextPrintLastApiError
 *
 * Print the details of the error set by the last flxActTranXXX function called by this thread. 
 ***************************************************************************************************/
void flxCtuContextPrintLastApiError(FlxCtuContext* pContext)
{
	FlxActTranResult		resultCode		= flxActTransactionGetResult();
	FlxActTranFunctionId	functionId		= flxActTransactionGetFunctionId();
	const char*				pszLogString	= flxActTransactionGetLogString();

	flxCtuPrintError("API function %u failed, result %u\n"
					 "%s\n",
					 functionId,
					 resultCode,
					 flxCtuGetTextForResult(resultCode));

	/* If verbose, output the log string. */
	flxCtuPrintVerbose("Log string: %s\n", pszLogString);

	/* If CMD_LOG_ERRORS specified, write the error details to the FNP event log. */
	if (flxCtuContextHasCommand(pContext, CMD_LOG_ERRORS))
	{
		if (flxActTransactionLogError())
			flxCtuPrintMessage("Error logged.\n");
		else
			flxCtuPrintError("flxActTransactionLogError() failed.\n");
	}

	pContext->exitCode = flxCtuExitApiError;
}

/***************************************************************************************************
*		flxCtuContextPrintLastApiFlexError
*
* Print an error set by a flxActXXX function other than a flxActTranXXX().
* The last error is obtained from the top level API handle.
***************************************************************************************************/
void flxCtuContextPrintLastApiFlexError(FlxCtuContext* pContext)
{
	FlxActError flxError;
	flxActCommonHandleGetError(pContext->hApi, &flxError);
	flxCtuContextPrintApiFlexError(pContext, &flxError);
}

/***************************************************************************************************
*		flxCtuContextValidateCommands
*
* Additional validation needed after successful command parsing - command combinations etc.
***************************************************************************************************/
FlxActBool flxCtuContextValidateCommands(FlxCtuContext* pContext)
{
	const FlxCtuCommand* pCommand;
	const FlxCtuCommand* pCommand2;

	/* Rule: if CMD_TRANSACTION is present then... */
	if (flxCtuContextGetCommandByName(pContext, CMD_TRANSACTION, &pCommand))
	{
		/* ...either CMD_NEW or CMD_STORED must also be. */
		if ( !(flxCtuContextHasCommand(pContext, CMD_NEW) ||
			   flxCtuContextHasCommand(pContext, CMD_STORED)))
		{
			flxCtuCommandPrintError(pCommand, "requires -%s or -%s\n", CMD_NEW, CMD_STORED);
			return FLX_ACT_FALSE;
		}
		/* ...the server URI must be specified as the command value. */
		if ( !flxCtuCommandHasValue(pCommand) )
		{
			flxCtuCommandPrintError(pCommand, "server URI not specified.\n");
			return FLX_ACT_FALSE;
		}
	}

	/* Rule: if CMD_NEW is present then ... */
	if (flxCtuContextGetCommandByName(pContext, CMD_NEW, &pCommand))
	{
		/* Any main command must be CMD_TRANSACTION. */
		if (pContext->bIsMainCommandSet)
		{
			if ( !flxCtuContextHasCommand(pContext, CMD_TRANSACTION))
			{
				flxCtuCommandPrintError(pCommand, "and \"%s\" cannot both be present.\n", 
					pContext->pMainCommand->pszNameForMessages);
				return FLX_ACT_FALSE;
			}
		}
		else
		{
			/* No main command so set this as the main command. */
			pContext->bIsMainCommandSet = FLX_ACT_TRUE;
			pContext->pMainCommand	    = pCommand;
		}
	}

	/* Rule: if CMD_STORED is present then ... */
	if (flxCtuContextGetCommandByName(pContext, CMD_STORED, &pCommand))
	{
		/* Any main command must be CMD_TRANSACTION. */
		if (pContext->bIsMainCommandSet)
		{
			if ( !flxCtuContextHasCommand(pContext, CMD_TRANSACTION))
			{
				flxCtuCommandPrintError(pCommand, "and \"%s\" cannot both be present.\n", 
										pContext->pMainCommand->pszNameForMessages);
				return FLX_ACT_FALSE;
			}
		}
		else
		{
			/* No main command so set this as the main command. */
			pContext->bIsMainCommandSet = FLX_ACT_TRUE;
			pContext->pMainCommand	    = pCommand;
		}
	}

	/* Rule: if there are any actions then CMD_NEW must be present. */
	if (sGetFirstRequestAction(pContext, &pCommand))
	{
		if ( !flxCtuContextHasCommand(pContext, CMD_NEW))
		{
			flxCtuCommandPrintError(pCommand, "is only valid with \"-%s\"\n", CMD_NEW);
			return FLX_ACT_FALSE;
		}
	}

	/* Rule: CMD_REPAIR and CMD_REPAIR_ALL cannot both be present. */
	if (	flxCtuContextGetCommandByName(pContext, CMD_REPAIR, &pCommand)
		 && flxCtuContextGetCommandByName(pContext, CMD_REPAIR_ALL, &pCommand2))
	{
		flxCtuCommandPrintError(pCommand, "is not valid with %s\n", pCommand2->pszNameForMessages);
		return FLX_ACT_FALSE;
	}

	/* Rule: if CMD_ACTIVATE is present with attribute CMD_ATTR_FR then ... */
	if (	flxCtuContextGetCommandByName(pContext, CMD_ACTIVATE, &pCommand)
		 && flxCtuCommandHasAttribute(pCommand, CMD_ATTR_FR))
	{
		/* ... attribute CMD_ATTR_FID must also be present. */
		if ( !flxCtuCommandHasAttribute(pCommand, CMD_ATTR_FID))
		{
			flxCtuCommandPrintError(pCommand, 
									"fulfillmentRecord|fr is only valid with fulfillmentId|fid\n");
			return FLX_ACT_FALSE;
		}
	}

	/* Rule: if CMD_ACTIVATE is present with attribute CMD_ATTR_EXP_TYPE then ... */
	if (	flxCtuContextGetCommandByName(pContext, CMD_ACTIVATE, &pCommand)
		 && flxCtuCommandHasAttribute(pCommand, CMD_ATTR_EXP_TYPE))
	{
		/* ... attribute CMD_ATTR_EXP_VALUE must also be present. */
		if ( !flxCtuCommandHasAttribute(pCommand, CMD_ATTR_EXP_VALUE))
		{
			flxCtuCommandPrintError(pCommand, 
									"expirationType|expType is only valid with expiration|exp\n");
			return FLX_ACT_FALSE;
		}
	}

	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*		flxCtuGetTextForResult
*
* Return text string for some of the more likely/interesting result codes.
***************************************************************************************************/
const char* flxCtuGetTextForResult(FlxActTranResult result)
{
	switch (result)
	{
	case FLX_ACT_TRAN_ERR_NO_MEMORY:		return "No memory.";
	case FLX_ACT_TRAN_ERR_HANDLE:			return "Invalid handle.";
	case FLX_ACT_TRAN_ERR_TRANS_HANDLE:		return "Invalid Transaction handle";
	case FLX_ACT_TRAN_BAD_REQ_HANDLE:		return "Invalid Request handle";
	case FLX_ACT_TRAN_ERR_RSP_XML_HEADER:	return "XML header invalid/wrong XML namespace";
	case FLX_ACT_TRAN_ERR_RSP_XML_MISSING:	return "Mandatory XML element missing";
	case FLX_ACT_TRAN_ERR_RSP_NOT_COMPOSITE:return "Not a composite response";
	case FLX_ACT_TRAN_ERR_RSP_XML_VERSION:	return "Response XML version number not supported";
	case FLX_ACT_TRAN_ERR_RSP_XML_OTHER:	return "Response XML invalid";
	case FLX_ACT_TRAN_ERR_RSP_SIGNATURE_1:	return "First (or only) signature invalid";
	case FLX_ACT_TRAN_ERR_RSP_NO_REQUEST:	return "No stored request for this response";
	case FLX_ACT_TRAN_ERR_RSP_NOT_REQUESTER:return "Request generated on a different machine";
	case FLX_ACT_TRAN_ERR_RSP_SIGNATURE_2:	return "Second or subsequent signature invalid";
	case FLX_ACT_TRAN_ERR_RSP_ACTION_INFO:	return "Invalid or missing response action attribute";
	case FLX_ACT_TRAN_ERR_RSP_ACTION_FAILED:return "One or more response actions failed";
	case FLX_ACT_TRAN_ERR_RSP_SIG_VERSION:	return "Response XML signature version not supported";
	case FLX_ACT_TRAN_ERR_TS_LOAD:			return "Trusted storage load failed";
	case FLX_ACT_TRAN_ERR_TS_SAVE:			return "Trusted storage save failed";
	case FLX_ACT_TRAN_ERR_FR_DISABLED:		return "Fulfillment record is disabled";
	case FLX_ACT_TRAN_ERR_FR_UNTRUSTED:		return "Fulfillment record is not fully trusted";
	case FLX_ACT_TRAN_ERR_FR_TRUSTED:		return "Fulfillment record does not need repair";
	case FLX_ACT_TRAN_ERR_RSP_NOT_XML:		return "Not an XML document";

	case FLX_ACT_TRAN_ERR_RSP_REQUEST_NOT_PENDING:	return "Stored request state incorrect";

	default:								return "";
	}
}

/***************************************************************************************************
*		sGetFirstRequestAction
*
***************************************************************************************************/
static FlxActBool sGetFirstRequestAction(const FlxCtuContext*	pContext, 
										 const FlxCtuCommand**	ppCommand)
{
	size_t commandIndex;

	for (commandIndex = 0; commandIndex < pContext->commandCount; commandIndex++)
	{
		const FlxCtuCommand* pCommand = pContext->ppCommands[commandIndex];
		if ((pCommand->pDef->traits & flxCtuCommandTraitIsRequestAction) != 0)
		{
			if (ppCommand != NULL)
				*ppCommand = pCommand;
			return FLX_ACT_TRUE;
		}
	}
	return FLX_ACT_FALSE;
}

/***************************************************************************************************
*		sFindCommand
*
***************************************************************************************************/
FlxActBool sFindCommand(const FlxCtuContext*	pContext, 
						const char*				pszCommandName, 
						size_t*					pCommandIndex)
{
	size_t commandIndex;

	for (commandIndex = 0; commandIndex < pContext->commandCount; commandIndex++)
	{
		if (strcmp(pszCommandName, pContext->ppCommands[commandIndex]->pDef->pszName) == 0)
		{
			if (pCommandIndex != NULL)
				*pCommandIndex = commandIndex;
			return FLX_ACT_TRUE;
		}
	}
	return FLX_ACT_FALSE;
}

